/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * Copyright (C) 2006-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://openbluesky.java.net
 * SCM: https://bitbucket.org/tidalwave/openbluesky-src
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.persistence;

import javax.annotation.Nonnull;
import java.util.List;
import java.util.logging.Level;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.EntityTransaction;
import javax.persistence.NoResultException;
import javax.persistence.PersistenceException;
import javax.persistence.Query;
import org.hibernate.exception.LockAcquisitionException;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.util.Task;

// FIXME: depends on Hibernate's LockAcquisitionException

class RetryException extends RuntimeException
  {
    public RetryException (final Throwable cause)
      {
        super(cause);
      }
  }

/*******************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 ******************************************************************************/
public abstract class TxTask<T, E extends Throwable> extends Task<T, E>
  {
    private final static String CLASS = TxTask.class.getName();
    private final static Logger logger = Logger.getLogger(CLASS);
    
    private final static int MAX_RETRY_COUNT = 5;

    private final static ThreadLocal<EntityManager> EM_HOLDER = new ThreadLocal<EntityManager>();
    
    private final static EntityManagerFactory entityManagerFactory = Locator.find(Persistence.class).getEntityManagerFactory();

    /***************************************************************************
     *
     * 
     **************************************************************************/
    @Override
    public abstract T run()
      throws E;
    
    /***************************************************************************
     *
     * 
     **************************************************************************/
    @Nonnull 
    public static EntityManager getEntityManager()
      {
        final EntityManager entityManager = EM_HOLDER.get();
        
        if (entityManager == null)
          {
            throw new IllegalStateException("No EntityManager bound to the current thread. " +
                                            "You must execute the calling code from the body of a TxTask.");
          }
        
        return entityManager;
      }
    
    /***************************************************************************
     *
     * 
     **************************************************************************/
    public static <T> T execute (final @Nonnull Task<T, RuntimeException> task)
      {
        return execute(RuntimeException.class, task);
      }
    
    /***************************************************************************
     *
     * 
     **************************************************************************/
    public static <T, E extends Throwable> T execute (final @Nonnull Class<E> exceptionClass,
                                                      final @Nonnull Task<T, E> task)
      throws E
      {
        final String where = findCallerName();
        final boolean demarcateTransaction = (EM_HOLDER.get() == null);
        T result = null;
        
        boolean done = false;
        boolean rolledback = false;
        final long time = System.currentTimeMillis();
        
        for (int retry = 1; !done; retry++)
          {
            if (demarcateTransaction)
              {
                beginTransaction(where, entityManagerFactory);
              }

            try
              {
                try
                  {
                    result = task.run();
                    done = true;
                  }
                catch (RetryException e)
                  {
                    retry(retry, e);
                  }
                catch (PersistenceException e) 
                  {
                    //
                    // Anything that is not related to locks is a severe error and can't be handled.
                    //
                    if (!(e.getCause() instanceof LockAcquisitionException)) 
                      {
                        throw e;  
                      }
                    else
                      {
                        //
                        // Lock exceptions must be handled here if we are the TxTask that demarcated
                        // the transaction.
                        //
                        if (demarcateTransaction)
                          {
                            retry(retry, e);
                          }
                        //
                        // Otherwise we throw a specific exception that will be handled by the TxTask
                        // that demarcated the transaction.
                        //
                        else
                          {
                            logger.fine(">>>> deadlock detected, throwing RetryException...");
                            throw new RetryException(e); 
                          }
                      }
                  }
              }
            catch (Throwable t)
              {
                finishTransactionInError(where, demarcateTransaction, t);

                if (exceptionClass.isAssignableFrom(t.getClass()))
                  {
                    throw (E)t;
                  }  
                //
                // In particular it's important that RetryException is not wrapped.
                //
                throw (t instanceof RuntimeException) ? (RuntimeException)t : new RuntimeException(t);
              }
            
            if (demarcateTransaction)
              {
                rolledback = !finishTransaction(where);
              }
          }
        
        logger.fine(">>>> TxTask(%s) executed in %d msec", where, System.currentTimeMillis() - time);

        if (rolledback)
          {
            throw new PersistenceException("Transaction rolled back - something called setRollbackOnly()");  
          }
            
        return result;
      }
    
    /***************************************************************************
     *
     * 
     **************************************************************************/
    @SuppressWarnings("unchecked")
    public static <T> List<T> query (final @Nonnull String jpql, final @Nonnull Object ... parameters)
      {
        return (List<T>)createQuery(getEntityManager(), jpql, parameters).getResultList();
      }
    
    /***************************************************************************
     *
     * 
     **************************************************************************/
    @SuppressWarnings("unchecked")
    @Nonnull
    public static <T> T querySingle (final @Nonnull String jpql, final @Nonnull Object ... parameters)
      throws NotFoundException
      {
        try
          {
            return (T)createQuery(getEntityManager(), jpql, parameters).getSingleResult();
          }
        catch (NoResultException e)
          {
            throw new NotFoundException(e);
          }
      }
    
    /***************************************************************************
     *
     * 
     **************************************************************************/
    public static int update (final @Nonnull String jpql, final @Nonnull Object ... parameters)
      {
        return createQuery(getEntityManager(), jpql, parameters).executeUpdate();
      }
    
    /***************************************************************************
     *
     * Do not use this method. Exposed only for testing purposes.
     * 
     **************************************************************************/
    public static void beginTransaction (final @Nonnull String where,
                                         final @Nonnull EntityManagerFactory entityManagerFactory)
      {
        logger.fine(">>>> create EntityManager, begin transaction [%s]", where);
        final EntityManager entityManager = entityManagerFactory.createEntityManager();
        EM_HOLDER.set(entityManager);
        entityManager.getTransaction().begin();
      }
    
    /***************************************************************************
     *
     * Do not use this method. Exposed only for testing purposes.
     * 
     **************************************************************************/
    public static boolean finishTransaction (final @Nonnull String where)
      {
        try
          {
            final EntityTransaction transaction = getEntityManager().getTransaction();

            if (!transaction.getRollbackOnly())
              {
                logger.fine(">>>> commit transaction [%s]", where);
                transaction.commit();
                return true;
              }
            else
              {
                logger.fine(">>>> rollback transaction [%s]", where);
                transaction.rollback();
                return false;
              }
          }
        finally
          {
            closeEntityManager(where);
          }
      }
    
    /***************************************************************************
     *
     * 
     **************************************************************************/
    private static void retry (final int retry, final Exception e)
      throws Exception
      {
        if (retry < MAX_RETRY_COUNT)
          {
            logger.info("Concurrency fault detected at try #%d, retrying...", retry);
            // FIXME: maybe a sleep here would do...
          }
        else
          {
            logger.severe("Concurrency fault detected, no more retries available");
            throw (e instanceof RetryException) ? (Exception)e.getCause() : e;
          }
      }
    
    /***************************************************************************
     *
     * 
     **************************************************************************/
    private static void finishTransactionInError (final String where, final boolean demarcateTransaction, final Throwable t)
      {
        logger.warning("Exception during transactional task: %s", t);
        logger.throwing(CLASS, "finishTransactionInError()", t);

        try
          {
            final EntityTransaction transaction = getEntityManager().getTransaction();

            if (demarcateTransaction)
              {
                logger.fine(">>>> rollback transaction [%s] because %s", where, t);
                transaction.rollback();
              }
//            else 
//              {
//                logger.fine(">>>> transaction set as rollback only [%s] because %s", where, t);
//                transaction.setRollbackOnly();  
//              }
          }
        catch (Exception e)
          {
            logger.warning("Error while closing transaction [%s]", where);
            logger.throwing(CLASS, "execute()", e);
          }
        finally
          {
            if (demarcateTransaction)
              {
                closeEntityManager(where);
              }
          }
      }
    
    /***************************************************************************
     *
     * 
     **************************************************************************/
    private static void closeEntityManager (final String where)
      {
        final EntityManager entityManager = EM_HOLDER.get();
        EM_HOLDER.set(null);
        
        try
          {
            entityManager.close();
            logger.fine(">>>> closed EntityManager [%s]", where);
          }
        catch (Exception e)
          {
            logger.warning("Error while closing EntityManager [%s]", where);
            logger.throwing(CLASS, "closeEntityManager()", e);
          }
      }
    
    /***************************************************************************
     *
     * 
     **************************************************************************/
    private static Query createQuery (final EntityManager entityManager, final String jpql, final Object ... parameters)
      {
        final Query query = entityManager.createQuery(jpql);
        
        if (logger.isLoggable(Level.FINEST))
          {
            String tmp = jpql;
            
            for (int i = 0; i < parameters.length; i += 2)
              {
                final String name = (String)parameters[i];
                final Object value = parameters[i + 1];
                try
                  {
                    tmp = tmp.replaceAll(":" + name, (value == null) ? "null" : value.toString());
                  }
                catch (IllegalArgumentException e) // FIXME
                  {
                    logger.warning("Cannot format %s in %s", name, tmp);
                  }

              }
    
            logger.info("QUERY: %s", tmp);
          }  

        for (int i = 0; i < parameters.length; i += 2)
          {
            final String name = (String)parameters[i];
            final Object value = parameters[i + 1];
            query.setParameter(name, value); 
          }
        
        return query;
      }
    
    /***************************************************************************
     *
     * 
     **************************************************************************/
    private static String findCallerName() 
      {
        final Thread currentThread = Thread.currentThread();
        
        for (final StackTraceElement element : currentThread.getStackTrace()) 
          {
            if (element.getClassName().equals(Thread.class.getName()))
              {
                continue;
              }
            
            if (!element.getClassName().equals(TxTask.class.getName()))
              {
                return String.format("%s#%d:%s.%s():%d", currentThread.getName(), 
                                                         currentThread.getId(),
                                                         element.getClassName(), 
                                                         element.getMethodName(), 
                                                         element.getLineNumber());
              }
          }

        return "unknown";
      }
  }
